#Author - Brian Ekins (EkinsSolutions.com)
#Description - Cuts a selected body into slices and lays them out on a plane to 
#              allow the creation of tool paths to machine the individual slices.
#(C) Copyright 2018 by Ekins Solutions, LLC.
# Permission to use, copy, modify, and distribute this software in any form for any 
# purpose and without fee is hereby granted, provided that the above copyright notice
# appears in all copies and that both that copyright notice and the limited warranty
# and restricted rights notice below appear in all supporting documentation.
    
# EKINS SOLUTIONS LLC PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. EKINS SOLUTIONS LLC
# SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
# EKINS SOLUTIONS LLC DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.

import adsk.core, adsk.fusion, adsk.cam, traceback
import math, os

_appPath = os.path.dirname(os.path.abspath(__file__))
_app = adsk.core.Application.cast(None)
_ui = adsk.core.UserInterface.cast(None)
_handlers = []


# Event handler for the commandCreated event.
class SlicerCommandCreatedHandler(adsk.core.CommandCreatedEventHandler):
    def __init__(self):
        super().__init__()
    def notify(self, args):
        try:
            eventArgs = adsk.core.CommandCreatedEventArgs.cast(args)
            cmd = eventArgs.command
            inputs = cmd.commandInputs
            
            # Define default values.
            toolDia = 0.25 * 2.54
            boardWidth = 48 * 2.54
            boardHeight = 48 * 2.54
            sliceThickness = 0.75 * 2.54
            showNumbers = True
            
            # Read the default values.
            try:
                if os.path.isfile(os.path.join(_appPath, 'defaults.dat')):
                    file = open(os.path.join(_appPath, 'defaults.dat'))
                    data = file.readlines()[0]
                    file.close()
    
                    settings = data.split(';')
                    for setting in settings:
                        settingData = setting.split(':')
                        
                        if settingData[0] == 'toolDia':
                            toolDia = float(settingData[1])
                        elif settingData[0] == 'boardWidth':
                            boardWidth = float(settingData[1])
                        elif settingData[0] == 'boardHeight':
                            boardHeight = float(settingData[1])
                        elif settingData[0] == 'sliceThickness':
                            sliceThickness = float(settingData[1])
                        elif settingData[0] == 'showNumbers':
                            if settingData[1] == 'True':
                                showNumbers = True
                            else:
                                showNumbers = False
            except:
                pass
                            
            # Create the command inputs.
            bodySel = inputs.addSelectionInput('bodySelect', 'Body', 'Select the body to slice.')
            bodySel.addSelectionFilter('Bodies')
            bodySel.setSelectionLimits(1,1)
            
            planeSel = inputs.addSelectionInput('faceSelect', 'Bottom face', 'Select a face on the bottom of the body.')
            planeSel.addSelectionFilter('PlanarFaces')
            planeSel.setSelectionLimits(1,1)
    
            # Look for a parameter named "Thickness".
            thickness = 0.75 * 2.54  # Initialize the thickness to be 3/4".
            des = adsk.fusion.Design.cast(_app.activeProduct)
            param = des.allParameters.itemByName('Thickness')
            if param:
                thickness = param.value
            else:
                param = des.allParameters.itemByName('thickness')
                if param:
                    thickness = param.value
                                
            thicknessVal = inputs.addValueInput('thickness', 'Slice thickness', des.unitsManager.defaultLengthUnits, adsk.core.ValueInput.createByReal(sliceThickness))
    
            boardWidth = inputs.addValueInput('boardWidth', 'Board width', des.unitsManager.defaultLengthUnits, adsk.core.ValueInput.createByReal(boardWidth))        
            boardHeight = inputs.addValueInput('boardHeight', 'Board height', des.unitsManager.defaultLengthUnits, adsk.core.ValueInput.createByReal(boardHeight))
            
            layoutOffset = inputs.addValueInput('layoutOffset', 'Tool diameter', des.unitsManager.defaultLengthUnits, adsk.core.ValueInput.createByReal(toolDia))
            
            addNumbers = inputs.addBoolValueInput('addNumbers', 'Number layers', True, '', showNumbers )
            
            # Connect to command related events.
            onExecute = SlicerCommandExecuteHandler()
            eventArgs.command.execute.add(onExecute)
            _handlers.append(onExecute)
            
            onInputChanged = SlicerCommandInputChangedHandler()
            eventArgs.command.inputChanged.add(onInputChanged)
            _handlers.append(onInputChanged)            

            onValidateInputs = SlicerCommandValidateInputsHandler()
            eventArgs.command.validateInputs.add(onValidateInputs)
            _handlers.append(onValidateInputs)
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


# Event handler for the inputChanged event.
class SlicerCommandInputChangedHandler(adsk.core.InputChangedEventHandler):
    def __init__(self):
        super().__init__()
    def notify(self, args):
        try:
            eventArgs = adsk.core.InputChangedEventArgs.cast(args)
            inputs = eventArgs.inputs

            if eventArgs.input.id == 'bodySelect':
                bodyInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('bodySelect'))
                if bodyInput.selectionCount == 1:
                    faceInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('faceSelect'))
                    faceInput.hasFocus = True
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))
        

# Event handler for the validateInputs event.
class SlicerCommandValidateInputsHandler(adsk.core.ValidateInputsEventHandler):
    def __init__(self):
        super().__init__()
    def notify(self, args):
        try:
            eventArgs = adsk.core.ValidateInputsEventArgs.cast(args)
            inputs = eventArgs.inputs
    
            # Get the inputs.
            bodyInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('bodySelect'))
            if bodyInput.selectionCount < 1:
                eventArgs.areInputsValid = False
                return
                
            faceInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('faceSelect'))
            if faceInput.selectionCount < 1:
                eventArgs.areInputsValid = False
                return
            
            thicknessInput = adsk.core.ValueCommandInput.cast(inputs.itemById('thickness'))
            if thicknessInput.value <= 0:
                eventArgs.areInputsValid = False
                return
            
            widthInput = adsk.core.ValueCommandInput.cast(inputs.itemById('boardWidth'))
            if widthInput.value <= 0:
                eventArgs.areInputsValid = False
                return
    
            heightInput = adsk.core.ValueCommandInput.cast(inputs.itemById('boardHeight'))
            if heightInput.value <= 0:
                eventArgs.areInputsValid = False
                return
            
            layoutOffsetInput = adsk.core.ValueCommandInput.cast(inputs.itemById('layoutOffset'))
            if layoutOffsetInput.value <= 0:
                eventArgs.areInputsValid = False
                return
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


class SlicerCommandExecuteHandler(adsk.core.CommandEventHandler):
    def __init__(self):
        super().__init__()
    def notify(self, args):
        try:
            eventArgs = adsk.core.CommandEventArgs.cast(args)
            inputs = eventArgs.command.commandInputs
            
            # Get the inputs.
            bodyInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('bodySelect'))
            body = bodyInput.selection(0).entity
            
            faceInput = adsk.core.SelectionCommandInput.cast(inputs.itemById('faceSelect'))
            face = faceInput.selection(0).entity
            
            thicknessInput = adsk.core.ValueCommandInput.cast(inputs.itemById('thickness'))
            thickness = thicknessInput.value
            
            widthInput = adsk.core.ValueCommandInput.cast(inputs.itemById('boardWidth'))
            width = widthInput.value
    
            heightInput = adsk.core.ValueCommandInput.cast(inputs.itemById('boardHeight'))
            height = heightInput.value
            
            layoutOffsetInput = adsk.core.ValueCommandInput.cast(inputs.itemById('layoutOffset'))
            layoutOffset = layoutOffsetInput.value
            
            addNumbersInput = adsk.core.BoolValueCommandInput.cast(inputs.itemById('addNumbers'))
            addNumbers = addNumbersInput.value
                        
            # Save the current settings as the new default values.            
            with open(os.path.join(_appPath, 'defaults.dat'), 'w') as file:
                file.write('toolDia:' + str(layoutOffset) + ';boardWidth:' + str(width) + ';boardHeight:' + str(height) + ';sliceThickness:' + str(thickness) + ';showNumbers:' + str(addNumbers))   
    
            createSlices(body, face, thickness, width, height, layoutOffset, addNumbers)            
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))
                
 
def createSlices(body, bottomFace, thickness, width, height, layoutOffset, addNumbers):
    try:
        # Get the current last node in the timeline.
        des = adsk.fusion.Design.cast(_app.activeProduct)
        timeline = adsk.fusion.Timeline.cast(des.timeline)
        firstNodeIndex = timeline.count

        # Slice the body.
        (occs, slicesOcc, startNode) = sliceBody(des, body, bottomFace, thickness)
        
        # Create a copy of the slices Occ to have a sliced version of the 3D model.
        newOcc = des.rootComponent.occurrences.addExistingComponent(slicesOcc.component, adsk.core.Matrix3D.create())
        
        if occs:
            arrangeSlices(des, slicesOcc, width, height, thickness, layoutOffset, occs, addNumbers)

        lastNode = des.timeline.item(timeline.count - 1)
        grp = timeline.timelineGroups.add(firstNodeIndex, lastNode.index)
        grp.name = 'Slices'    
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def arrangeSlices(des, slicesOcc, width, height, thickness, toolDia, occs, addNumbers):
    try:
        # Draw an occurrence that represents the stock panel.
        panelOcc = slicesOcc.component.occurrences.addNewComponent(adsk.core.Matrix3D.create())
        panelOccProxy = panelOcc.createForAssemblyContext(slicesOcc)
        panelOccProxy.isGrounded = True
        panelComp = panelOcc.component
        panelComp.name = 'Stock Board'
        sk = panelComp.sketches.add(panelComp.xYConstructionPlane)
        sk.sketchCurves.sketchLines.addTwoPointRectangle(adsk.core.Point3D.create(0,0,0), adsk.core.Point3D.create(width, height, 0))
        
        panelExtrude = panelComp.features.extrudeFeatures.addSimple(sk.profiles.item(0), adsk.core.ValueInput.createByReal(-thickness), adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
        panelFace = panelExtrude.startFaces.item(0)

        # Iterate through all of the occurrences.
        upVec = adsk.core.Vector3D.create(0,0,1)
        downVec = adsk.core.Vector3D.create(0,0,-1)
        currentX = toolDia
        currentY = toolDia
        maxY = 0
        occCount = 0
        currentColumn = 0
        currentRow = 0
        for occ in occs:
            body = adsk.fusion.BRepBody.cast(occ.bRepBodies.item(0))
            bodyXSize = body.boundingBox.maxPoint.x - body.boundingBox.minPoint.x
            bodyYSize = body.boundingBox.maxPoint.y - body.boundingBox.minPoint.y
                            
            # Check to see if it's extended beyone the board and start a new row.
            if currentX + bodyXSize > width:
                currentX = toolDia
                currentY = maxY + (toolDia * 2)
                currentRow += 1
                currentColumn = 0

            newCorner = adsk.core.Point3D.create(currentX, currentY, 0)
            
            # Find the top and bottom faces of the current body. 
            topFace = None
            bottomFace = None
            face = adsk.fusion.BRepFace.cast(None)
            for face in body.faces:
                if face.geometry.objectType == adsk.core.Plane.classType():
                    (retVal, normal) = face.evaluator.getNormalAtPoint(face.pointOnFace) 
                    if normal.angleTo(upVec) < _app.vectorAngleTolerance:
                        topFace = face
                    elif normal.angleTo(downVec) < _app.vectorAngleTolerance:
                        bottomFace = face
                        
            sketchFace = adsk.fusion.BRepFace.cast(None)
            if not topFace:
                sketchface = bottomFace
                orientedBottom = bottomFace
                isFlipped = False
            elif not bottomFace:
                sketchFace = topFace
                orientedBottom = topFace
                isFlipped = True
            else:
                if bottomFace.area > topFace.area:
                    sketchFace = topFace
                    orientedBottom = bottomFace
                    isFlipped = False
                else:
                    sketchFace = bottomFace
                    orientedBottom = topFace
                    isFlipped = True
            
            # Create a sketch and place text to label the placement.
            if addNumbers:
                comp = adsk.fusion.Component.cast(occ.component)
                sk = comp.sketches.addWithoutEdges(sketchFace)
                pnt = sk.modelToSketchSpace(sketchFace.centroid)
                occCount = occCount + 1
                txtInput = sk.sketchTexts.createInput(str(occCount), 2, pnt)
                sk.sketchTexts.add(txtInput)
    
            # Position the occurrence so the smaller of the top or bottom is on the top.
            if not isFlipped:
                existingCorner = adsk.core.Point3D.create(body.boundingBox.minPoint.x, body.boundingBox.minPoint.y, orientedBottom.pointOnFace.z)
                matrix = occ.transform

                moveTrans = existingCorner.vectorTo(newCorner)
                tempTrans = matrix.translation
                tempTrans.add(moveTrans)
                matrix.translation = tempTrans
                occ.transform = matrix
                des.snapshots.add()
            else:
                # Build existing coordinate system.
                existingOrigin = adsk.core.Point3D.create(body.boundingBox.maxPoint.x, body.boundingBox.minPoint.y, body.boundingBox.maxPoint.z)
                existingXDir = adsk.core.Vector3D.create(-1,0,0)
                existingYDir = adsk.core.Vector3D.create(0,1,0)
                existingZDir = adsk.core.Vector3D.create(0,0,-1)
                existingMatrix = adsk.core.Matrix3D.create()
                existingMatrix.setWithCoordinateSystem(existingOrigin, existingXDir, existingYDir, existingZDir)

                newOrigin = newCorner
                newXDir = adsk.core.Vector3D.create(1,0,0)
                newYDir = adsk.core.Vector3D.create(0,1,0)
                newZDir = adsk.core.Vector3D.create(0,0,1)
                newMatrix = adsk.core.Matrix3D.create()
                newMatrix.setWithCoordinateSystem(newOrigin, newXDir, newYDir, newZDir)
                
                existingMatrix.invert()
                existingMatrix.transformBy(newMatrix)
                
                occMatrix = occ.transform
                occMatrix.transformBy(existingMatrix)
                occ.transform = occMatrix
                des.snapshots.add()

            # Update the current position.            
            currentColumn += 1
            currentX += bodyXSize + (toolDia * 2)
            if currentY + bodyYSize > maxY:
                maxY = currentY + bodyYSize
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))
        
            

def sliceBody(des, body, bottomFace, sliceThickness):
    try:
        # Create a new component.
        slicesOcc = des.rootComponent.occurrences.addNewComponent(adsk.core.Matrix3D.create())
        startNode = slicesOcc.timelineObject
        slicesComp = adsk.fusion.Component.cast(slicesOcc.component)
        slicesComp.name = 'Slices'
        
        # Determine the number of cuts to be made.
        totalDist = body.boundingBox.maxPoint.z - body.boundingBox.minPoint.z
        cutCount = int(totalDist / sliceThickness)
        if cutCount * sliceThickness < totalDist:
            cutCount += 1
        
#        progDialog = _ui.createProgressDialog()
#        progDialog.isCancelButtonShown = False
#        progDialog.isBackgroundTranslucent = False
#        progDialog.show('Progress Dialog', 'Processing step %v of %m', 0, cutCount, 1)

        # Create a copy of the body.
        tempBRep = adsk.fusion.TemporaryBRepManager.get()
        tempBody = tempBRep.copy(body)
        
        # Transform the body so the selected face lies on the X-Y plane and it 
        # is in the -X, +Y quadrant.
        origin = bottomFace.pointOnFace
        (retVal, zVec) = bottomFace.evaluator.getNormalAtPoint(bottomFace.pointOnFace)
        zVec.scaleBy(-1)

        if zVec.angleTo(adsk.core.Vector3D.create(0,0,1)) < (math.pi * 0.25):
            xVec = adsk.core.Vector3D.create(1,0,0)
        elif zVec.angleTo(adsk.core.Vector3D.create(1,0,0)) < (math.pi * 0.25):
            xVec = adsk.core.Vector3D.create(0,1,0)
        else:
            xVec = adsk.core.Vector3D.create(0,0,1)

        yVec = zVec.crossProduct(xVec)
        xVec = yVec.crossProduct(zVec)
        xVec.normalize()
        yVec.normalize()
        trans = adsk.core.Matrix3D.create()
        trans.setWithCoordinateSystem(origin, xVec, yVec, zVec)
        transInv = trans.copy()
        transInv.invert()                   
        tempBRep.transform(tempBody, transInv)
        
        moveX = -(tempBody.boundingBox.maxPoint.x + ((tempBody.boundingBox.maxPoint.x - tempBody.boundingBox.minPoint.x) * 0.1))
        moveY = -(tempBody.boundingBox.minPoint.y)
        moveTrans = adsk.core.Matrix3D.create()
        moveTrans.translation = adsk.core.Vector3D.create(moveX, moveY, 0)
        tempBRep.transform(tempBody, moveTrans)
        
        # Define the coordinates of a box that encompasses the range in X-Y.
        bodyRange = tempBody.boundingBox
        centerXY = adsk.core.Point2D.create((bodyRange.minPoint.x + bodyRange.maxPoint.x)*0.5, (bodyRange.minPoint.y + bodyRange.maxPoint.y)*0.5)
        length = (bodyRange.maxPoint.x - bodyRange.minPoint.x) * 1.1
        width = (bodyRange.maxPoint.y - bodyRange.minPoint.y) * 1.1
               
        # Create a block slice for each section and intersect it with the body.
        sliceBodies = []
        for i in range(0, cutCount):
            # Calculate the dimensions of a box that fills the slizce area.            
            if i == 0:
                center = adsk.core.Point3D.create(centerXY.x, centerXY.y, bodyRange.minPoint.z + (sliceThickness * 0.5) - 0.05)
                box = adsk.core.OrientedBoundingBox3D.create(center, adsk.core.Vector3D.create(1,0,0), adsk.core.Vector3D.create(0,1,0), length, width, sliceThickness + 0.1)
            elif i == cutCount - 1:
                center = adsk.core.Point3D.create(centerXY.x, centerXY.y, bodyRange.minPoint.z + (sliceThickness * i) + (sliceThickness * 0.5) + 0.05)
                box = adsk.core.OrientedBoundingBox3D.create(center, adsk.core.Vector3D.create(1,0,0), adsk.core.Vector3D.create(0,1,0), length, width, sliceThickness + 0.1)
            else:
                center = adsk.core.Point3D.create(centerXY.x, centerXY.y, bodyRange.minPoint.z + (sliceThickness * i) + (sliceThickness * 0.5))
                box = adsk.core.OrientedBoundingBox3D.create(center, adsk.core.Vector3D.create(1,0,0), adsk.core.Vector3D.create(0,1,0), length, width, sliceThickness)

            # Create a body representing the box.
            sliceBody = tempBRep.createBox(box)

            # Intersect the box with the selected body to create the slice.                    
            tempBRep.booleanOperation(sliceBody, tempBody, adsk.fusion.BooleanTypes.IntersectionBooleanType)            

            # Check to see if there is more than one lump in the result body and create single lump bodies if there is more than one lump.
            if sliceBody.lumps.count > 1:
                # Create a single body for each lump found.
                for i in range(sliceBody.lumps.count):
                    bodyCopy = tempBRep.copy(sliceBody)
                    faces = []
                    lumpFaces = bodyCopy.lumps.item(i).faces
                    for face in lumpFaces:
                        faces.append(face)
                        
                    tempBRep.deleteFaces(faces, False)
                    sliceBodies.append(bodyCopy)
            else:
                sliceBodies.append(sliceBody)

        # Create real bodies for each section body.
        baseFeature = slicesComp.features.baseFeatures.add()
        baseFeature.startEdit()
        
        cnt = 0
        for body in sliceBodies:
            newBody = slicesComp.bRepBodies.add(body, baseFeature)
            
            cnt += 1
            newBody.name = 'Slice Body ' + str(cnt)
            
        baseFeature.finishEdit()
        
        # Create a new component for each body.
        occList = []
        for i in range(baseFeature.bodies.count):
            body = baseFeature.bodies.item(0)
            body.name = 'Slice ' + str(i+1)

            bodyOcc = slicesComp.occurrences.addNewComponent(adsk.core.Matrix3D.create())
            bodyComp = bodyOcc.component
            bodyComp.name = 'Slice ' + str(i+1)
            body.moveToComponent(bodyOcc)            

            bodyOccProxy = bodyOcc.createForAssemblyContext(slicesOcc)
            
            occList.append(bodyOccProxy)

        timeline = adsk.fusion.Timeline.cast(des.timeline)
        lastNode = des.timeline.item(timeline.count - 1)
        
        return (occList, slicesOcc, lastNode)

#        progDialog.hide()
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

        
def run(context):
    try:
        global _app, _ui
        _app = adsk.core.Application.get()
        _ui  = _app.userInterface
        
        # Create the command definition.
        sliceCmdDef = _ui.commandDefinitions.addButtonDefinition('ekinsSliceItUp', 
                                                                 'Slice It Up',
                                                                 'Slices the selected body and lays out the slices.\n\nThe resulting slices can be used to manufacture the individual slices.',
                                                                 'resources/SliceItUp')                                                       
        sliceCmdDef.toolClipFilename = 'resources/toolClip.png'

        # Connect to the command created event.
        onCommandCreated = SlicerCommandCreatedHandler()
        sliceCmdDef.commandCreated.add(onCommandCreated)
        _handlers.append(onCommandCreated)        

        # Add the command to the CREATE panel below the "Boundary Fill command.
        createPanel = _ui.allToolbarPanels.itemById('SolidCreatePanel')
        createPanel.controls.addCommand(sliceCmdDef, 'SurfaceSculpt', False)
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def stop(context):
    try:
        # Clean up the UI.
        sliceCmdDef = _ui.commandDefinitions.itemById('ekinsSliceItUp')
        if sliceCmdDef:
            sliceCmdDef.deleteMe()
        
        createPanel = _ui.allToolbarPanels.itemById('SolidCreatePanel')
        if createPanel:
            sliceControl = createPanel.controls.itemById('ekinsSliceItUp')
            if sliceControl:
                sliceControl.deleteMe()
    except:
        _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))
